﻿// Copyright (c) Microsoft. All rights reserved.

using System;
using Microsoft.SemanticKernel.Diagnostics;

namespace Microsoft.SemanticKernel.AI;

#pragma warning disable CA1032 // Implement standard exception constructors

/// <summary>
/// Exception thrown for errors related to AI logic.
/// </summary>
public class AIException : SKException
{
    /// <summary>
    /// Initializes a new instance of the <see cref="AIException"/> class with a provided error code.
    /// </summary>
    /// <param name="errorCode">The error code.</param>
    public AIException(ErrorCodes errorCode)
        : this(errorCode, message: null, detail: null, innerException: null)
    {
    }

    /// <summary>
    /// Initializes a new instance of the <see cref="AIException"/> class with a provided error code and message.
    /// </summary>
    /// <param name="errorCode">The error code.</param>
    /// <param name="message">The exception message.</param>
    public AIException(ErrorCodes errorCode, string? message)
        : this(errorCode, message, detail: null, innerException: null)
    {
    }

    /// <summary>
    /// Initializes a new instance of the <see cref="AIException"/> class with a provided error code, message, and inner exception.
    /// </summary>
    /// <param name="errorCode">The error code.</param>
    /// <param name="message">A string that describes the error.</param>
    /// <param name="innerException">The exception that is the cause of the current exception.</param>
    public AIException(ErrorCodes errorCode, string? message, Exception? innerException)
        : this(errorCode, message, detail: null, innerException)
    {
    }

    /// <summary>
    /// Initializes a new instance of the <see cref="AIException"/> class with a provided error code, message, and inner exception.
    /// </summary>
    /// <param name="errorCode">The error code.</param>
    /// <param name="message">A string that describes the error.</param>
    /// <param name="detail">A string that provides additional details about the error.</param>
    public AIException(ErrorCodes errorCode, string? message, string? detail)
        : this(errorCode, message, detail, innerException: null)
    {
    }

    /// <summary>
    /// Initializes a new instance of the <see cref="AIException"/> class with a provided error code, message, additional details, and inner exception.
    /// </summary>
    /// <param name="errorCode">The error code.</param>
    /// <param name="message">A string that describes the error.</param>
    /// <param name="detail">A string that provides additional details about the error.</param>
    /// <param name="innerException">The exception that is the cause of the current exception.</param>
    public AIException(ErrorCodes errorCode, string? message, string? detail, Exception? innerException)
        : base(GetDefaultMessage(errorCode, message), innerException)
    {
        this.ErrorCode = errorCode;
        this.Detail = detail;
    }

    /// <summary>
    /// Gets the error code for this exception.
    /// </summary>
    public ErrorCodes ErrorCode { get; }

    /// <summary>
    /// Gets the extended details for this exception.
    /// </summary>
    public string? Detail { get; }

    /// <summary>Translate the error code into a default message.</summary>
    private static string GetDefaultMessage(ErrorCodes errorCode, string? message)
    {
        string description = errorCode switch
        {
            ErrorCodes.NoResponse => "No response",
            ErrorCodes.AccessDenied => "Access denied",
            ErrorCodes.InvalidRequest => "Invalid request",
            ErrorCodes.InvalidResponseContent => "Invalid response content",
            ErrorCodes.Throttling => "Throttling",
            ErrorCodes.RequestTimeout => "Request timeout",
            ErrorCodes.ServiceError => "Service error",
            ErrorCodes.ModelNotAvailable => "Model not available",
            ErrorCodes.InvalidConfiguration => "Invalid configuration",
            ErrorCodes.FunctionTypeNotSupported => "Function type not supported",
            _ => $"Unknown error ({errorCode:G})",
        };

        return message is not null ? $"{description}: {message}" : description;
    }

    /// <summary>
    /// Possible error codes for exceptions
    /// </summary>
    public enum ErrorCodes
    {
        /// <summary>
        /// Unknown error.
        /// </summary>
        UnknownError = -1,

        /// <summary>
        /// No response.
        /// </summary>
        NoResponse,

        /// <summary>
        /// Access is denied.
        /// </summary>
        AccessDenied,

        /// <summary>
        /// The request was invalid.
        /// </summary>
        InvalidRequest,

        /// <summary>
        /// The content of the response was invalid.
        /// </summary>
        InvalidResponseContent,

        /// <summary>
        /// The request was throttled.
        /// </summary>
        Throttling,

        /// <summary>
        /// The request timed out.
        /// </summary>
        RequestTimeout,

        /// <summary>
        /// There was an error in the service.
        /// </summary>
        ServiceError,

        /// <summary>
        /// The requested model is not available.
        /// </summary>
        ModelNotAvailable,

        /// <summary>
        /// The supplied configuration was invalid.
        /// </summary>
        InvalidConfiguration,

        /// <summary>
        /// The function is not supported.
        /// </summary>
        FunctionTypeNotSupported,
    }
}
