﻿// Copyright (c) Microsoft. All rights reserved.

using Microsoft.SemanticKernel.TemplateEngine;
using Microsoft.SemanticKernel.TemplateEngine.Blocks;
using Xunit;

namespace SemanticKernel.UnitTests.TemplateEngine;

public class CodeTokenizerTests
{
    private readonly CodeTokenizer _target;

    public CodeTokenizerTests()
    {
        this._target = new CodeTokenizer();
    }

    [Fact]
    public void ItParsesEmptyText()
    {
        // Act +  Assert
        Assert.Empty(this._target.Tokenize(null));
        Assert.Empty(this._target.Tokenize(""));
        Assert.Empty(this._target.Tokenize(" "));
        Assert.Empty(this._target.Tokenize(" \n "));
    }

    [Theory]
    [InlineData("$", "$")]
    [InlineData(" $ ", "$")]
    [InlineData("$foo", "$foo")]
    [InlineData("$foo ", "$foo")]
    [InlineData(" $foo", "$foo")]
    [InlineData(" $bar ", "$bar")]
    public void ItParsesVarBlocks(string template, string content)
    {
        // Act
        var blocks = this._target.Tokenize(template);

        // Assert
        Assert.Single(blocks);
        Assert.Equal(content, blocks[0].Content);
        Assert.Equal(BlockTypes.Variable, blocks[0].Type);
    }

    [Theory]
    [InlineData("'", "'")]
    [InlineData(" \" ", "\"")]
    [InlineData("'foo'", "'foo'")]
    [InlineData("'foo' ", "'foo'")]
    [InlineData(" 'foo'", "'foo'")]
    [InlineData(" \"bar\" ", "\"bar\"")]
    public void ItParsesValBlocks(string template, string content)
    {
        // Act
        var blocks = this._target.Tokenize(template);

        // Assert
        Assert.Single(blocks);
        Assert.Equal(content, blocks[0].Content);
        Assert.Equal(BlockTypes.Value, blocks[0].Type);
    }

    [Theory]
    [InlineData("f", "f")]
    [InlineData(" x ", "x")]
    [InlineData("foo", "foo")]
    [InlineData("fo.o ", "fo.o")]
    [InlineData(" f.oo", "f.oo")]
    [InlineData(" bar ", "bar")]
    public void ItParsesFunctionIdBlocks(string template, string content)
    {
        // Act
        var blocks = this._target.Tokenize(template);

        // Assert
        Assert.Single(blocks);
        Assert.Equal(content, blocks[0].Content);
        Assert.Equal(BlockTypes.FunctionId, blocks[0].Type);
    }

    [Fact]
    public void ItParsesFunctionCalls()
    {
        // Arrange
        var template1 = "x.y $foo";
        var template2 = "xy $foo";
        var template3 = "xy '$value'";

        // Act
        var blocks1 = this._target.Tokenize(template1);
        var blocks2 = this._target.Tokenize(template2);
        var blocks3 = this._target.Tokenize(template3);

        // Assert
        Assert.Equal(2, blocks1.Count);
        Assert.Equal(2, blocks2.Count);
        Assert.Equal(2, blocks3.Count);

        Assert.Equal("x.y", blocks1[0].Content);
        Assert.Equal("xy", blocks2[0].Content);
        Assert.Equal("xy", blocks3[0].Content);

        Assert.Equal(BlockTypes.FunctionId, blocks1[0].Type);
        Assert.Equal(BlockTypes.FunctionId, blocks2[0].Type);
        Assert.Equal(BlockTypes.FunctionId, blocks3[0].Type);

        Assert.Equal("$foo", blocks1[1].Content);
        Assert.Equal("$foo", blocks2[1].Content);
        Assert.Equal("'$value'", blocks3[1].Content);

        Assert.Equal(BlockTypes.Variable, blocks1[1].Type);
        Assert.Equal(BlockTypes.Variable, blocks2[1].Type);
        Assert.Equal(BlockTypes.Value, blocks3[1].Type);
    }

    [Fact]
    public void ItSupportsEscaping()
    {
        // Arrange
        var template = "func 'f\\'oo'";

        // Act
        var blocks = this._target.Tokenize(template);

        // Assert
        Assert.Equal(2, blocks.Count);
        Assert.Equal("func", blocks[0].Content);
        Assert.Equal("'f\'oo'", blocks[1].Content);
    }

    [Fact]
    public void ItThrowsWhenSeparatorsAreMissing()
    {
        // Arrange
        var template1 = @"call 'f\\'xy'";
        var template2 = @"call 'f\\'x";

        // Act
        Assert.Throws<TemplateException>(() => this._target.Tokenize(template1));
        Assert.Throws<TemplateException>(() => this._target.Tokenize(template2));
    }
}
