﻿// Copyright (c) Microsoft. All rights reserved.

using System;
using System.IO;
using System.Threading.Tasks;
using Microsoft.Extensions.Logging;
using Microsoft.Extensions.Logging.Abstractions;
using Microsoft.SemanticKernel.Orchestration;
using Microsoft.SemanticKernel.SkillDefinition;
using Microsoft.SemanticKernel.Skills.Document.FileSystem;

namespace Microsoft.SemanticKernel.Skills.Document;

//**********************************************************************************************************************
// EXAMPLE USAGE
// Option #1: as a standalone C# function
//
// DocumentSkill documentSkill = new(new WordDocumentConnector(), new LocalDriveConnector());
// string filePath = "PATH_TO_DOCX_FILE.docx";
// string text = await documentSkill.ReadTextAsync(filePath);
// Console.WriteLine(text);
//
//
// Option #2: with the Semantic Kernel
//
// DocumentSkill documentSkill = new(new WordDocumentConnector(), new LocalDriveConnector());
// string filePath = "PATH_TO_DOCX_FILE.docx";
// ISemanticKernel kernel = SemanticKernel.Build();
// var result = await kernel.RunAsync(
//      filePath,
//      documentSkill.ReadTextAsync);
// Console.WriteLine(result);
//**********************************************************************************************************************

/// <summary>
/// Skill for interacting with documents (e.g. Microsoft Word)
/// </summary>
public class DocumentSkill
{
    /// <summary>
    /// <see cref="ContextVariables"/> parameter names.
    /// </summary>
    public static class Parameters
    {
        /// <summary>
        /// Document file path.
        /// </summary>
        public const string FilePath = "filePath";
    }

    private readonly IDocumentConnector _documentConnector;
    private readonly IFileSystemConnector _fileSystemConnector;
    private readonly ILogger<DocumentSkill> _logger;

    /// <summary>
    /// Initializes a new instance of the <see cref="DocumentSkill"/> class.
    /// </summary>
    /// <param name="documentConnector">Document connector</param>
    /// <param name="fileSystemConnector">File system connector</param>
    /// <param name="logger">Optional logger</param>
    public DocumentSkill(IDocumentConnector documentConnector, IFileSystemConnector fileSystemConnector, ILogger<DocumentSkill>? logger = null)
    {
        this._documentConnector = documentConnector ?? throw new ArgumentNullException(nameof(documentConnector));
        this._fileSystemConnector = fileSystemConnector ?? throw new ArgumentNullException(nameof(fileSystemConnector));
        this._logger = logger ?? new NullLogger<DocumentSkill>();
    }

    /// <summary>
    /// Read all text from a document, using <see cref="ContextVariables.Input"/> as the file path.
    /// </summary>
    [SKFunction("Read all text from a document")]
    [SKFunctionInput(Description = "Path to the file to read")]
    public async Task<string> ReadTextAsync(string filePath, SKContext context)
    {
        this._logger.LogInformation("Reading text from {0}", filePath);
        using var stream = await this._fileSystemConnector.GetFileContentStreamAsync(filePath, context.CancellationToken).ConfigureAwait(false);
        return this._documentConnector.ReadText(stream);
    }

    /// <summary>
    /// Append the text in <see cref="ContextVariables.Input"/> to a document. If the document doesn't exist, it will be created.
    /// </summary>
    [SKFunction("Append text to a document. If the document doesn't exist, it will be created.")]
    [SKFunctionInput(Description = "Text to append")]
    [SKFunctionContextParameter(Name = Parameters.FilePath, Description = "Destination file path")]
    public async Task AppendTextAsync(string text, SKContext context)
    {
        if (!context.Variables.Get(Parameters.FilePath, out string filePath))
        {
            context.Fail($"Missing variable {Parameters.FilePath}.");
            return;
        }

        // If the document already exists, open it. If not, create it.
        if (await this._fileSystemConnector.FileExistsAsync(filePath).ConfigureAwait(false))
        {
            this._logger.LogInformation("Writing text to file {0}", filePath);
            using Stream stream = await this._fileSystemConnector.GetWriteableFileStreamAsync(filePath, context.CancellationToken).ConfigureAwait(false);
            this._documentConnector.AppendText(stream, text);
        }
        else
        {
            this._logger.LogInformation("File does not exist. Creating file at {0}", filePath);
            using Stream stream = await this._fileSystemConnector.CreateFileAsync(filePath).ConfigureAwait(false);
            this._documentConnector.Initialize(stream);

            this._logger.LogInformation("Writing text to {0}", filePath);
            this._documentConnector.AppendText(stream, text);
        }
    }
}
